# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

import couch
from tornado import httpclient


class ValidationError(Exception):
    pass


class Unauthorized(Exception):
    pass


class NotFound(couch.NotFound):
    def __init__(self):
        http_error = httpclient.HTTPError(404)
        couch.NotFound.__init__(self, http_error)
